#define HASHTABLE_KEEP_DEFINITIONS

#include "hashtable.h"

#include <stdlib.h>

#define HASHTABLE_DEBUG(ARGS)		/* fprintf ARGS */

#if HASHTABLE_VERBOSE > 0
# include <stdio.h>
# define HASHTABLE_INFO(ARGS)		fprintf ARGS
#else
# define HASHTABLE_INFO(ARGS)
#endif

#if HASHTABLE_VERBOSE > 1
# define HASHTABLE_DEBUG(ARGS)		fprintf ARGS
#else
# define HASHTABLE_DEBUG(ARGS)
#endif

#ifndef HASHTABLE_KEY_COPY
# include <string.h>
# define HASHTABLE_KEY_COPY		strdup
#endif

#ifndef HASHTABLE_KEY_COMPARE
# include <string.h>
# define HASHTABLE_KEY_COMPARE		!strcmp
#endif

#ifndef HASHTABLE_FRACTION
# define HASHTABLE_FRACTION		1/2
#endif

#ifndef HASHTABLE_FACTOR
# define HASHTABLE_FACTOR		3
#endif

#ifndef HASHTABLE_HASH_FUNCTION
# define HASHTABLE_HASH_FUNCTION HASHTABLE(hash_function)
  static inline unsigned HASHTABLE_HASH_FUNCTION(HASHTABLE_KEY_T key, unsigned len)
  {
      unsigned h= 1315423911;
      while (len--) h ^= (h << 5) + (h >> 2) + *key++;
      return h;
  }
#endif

HASHTABLE_T *HASHTABLE(initialise)(HASHTABLE_T *h, unsigned capacity)
{
    if (h) {
	if (capacity < 2) capacity= 2;
	h->hashtable_capacity= capacity;
	h->hashtable_lastfree= capacity;
	h->hashtable_size= capacity * 86 / 100;
	h->hashtable_tally= 0;
	h->hashtable_entries= calloc(capacity, sizeof(HASHTABLE_ENTRY_T *));
	if (!h->hashtable_entries) h= 0;
    }
    return h;
}

HASHTABLE_T *HASHTABLE(new)(unsigned hashtable_capacity)
{
    HASHTABLE_T *h= malloc(sizeof(HASHTABLE_T));
    if (h) h= HASHTABLE(initialise)(h, hashtable_capacity);
    return h;
}

HASHTABLE_T *HASHTABLE(release)(HASHTABLE_T *h)
{
    if (h->hashtable_entries) {
	int i;
	for (i= 0;  i < h->hashtable_capacity;  ++i)
	    if (h->hashtable_entries[i]) {
		free(h->hashtable_entries[i]->hashtable_entry_key);
		free(h->hashtable_entries[i]);
	    }
	free(h->hashtable_entries);
	h->hashtable_entries= 0;
    }
    h->hashtable_capacity= 0;
    h->hashtable_size= 0;
    return h;
}

void HASHTABLE(delete)(HASHTABLE_T *h)
{
    free(HASHTABLE(release)(h));
}

HASHTABLE_ENTRY_T *HASHTABLE(make_entry)(HASHTABLE_KEY_T key, unsigned len)
{
    HASHTABLE_ENTRY_T *entry= calloc(1, sizeof(HASHTABLE_ENTRY_T));
    if (entry) {
	entry->hashtable_entry_key= HASHTABLE_KEY_COPY(key);
	entry->hashtable_entry_length= len;
	entry->hashtable_entry_next= 0;
    }
    return entry;
}

unsigned HASHTABLE(grow)(HASHTABLE_T *h);

HASHTABLE_ENTRY_T *HASHTABLE(insert_entry)(HASHTABLE_T *t, HASHTABLE_ENTRY_T *entry)
{
    HASHTABLE_ENTRY_T **entries;
    unsigned h;

    if (t->hashtable_tally > t->hashtable_capacity * HASHTABLE_FRACTION)
	if (!HASHTABLE(grow)(t))
	    return 0;

    entries= t->hashtable_entries;
    h= HASHTABLE_HASH_FUNCTION(entry->hashtable_entry_key, entry->hashtable_entry_length) % t->hashtable_size;
    entry->hashtable_entry_next= 0;

    HASHTABLE_DEBUG((stderr, "INSERT ENTRY\t%s\t%08x %p\n", entry->hashtable_entry_key, h, entry));

    if (!entries[h])				/* new chain */
	entries[h]= entry;
    else {
	HASHTABLE_ENTRY_T *chain;
	int last= t->hashtable_lastfree;
	while (--last >= 0 && entries[last])	/* find empty bucket */
	    ;
	t->hashtable_lastfree= last;

	if (last == -1)				/* table full -- should never happen */
	    return 0;

	entries[last]= entry;

	chain= entries[h];
	while (chain->hashtable_entry_next)
	    chain= chain->hashtable_entry_next;
	chain->hashtable_entry_next= entry;	/* append to chain */
    }
    t->hashtable_tally++;
    return entry;
}

HASHTABLE_ENTRY_T *HASHTABLE(insert)(HASHTABLE_T *t, HASHTABLE_KEY_T key, unsigned l)
{
    HASHTABLE_ENTRY_T *entry= HASHTABLE(make_entry)(key, l);
    if (entry) entry= HASHTABLE(insert_entry)(t, entry);
    return entry;
}

unsigned HASHTABLE(grow)(HASHTABLE_T *h)
{
    HASHTABLE_T *big= HASHTABLE(new)(h->hashtable_capacity * HASHTABLE_FACTOR);
    unsigned i;
    if (!big) return 0;
    HASHTABLE_INFO((stderr, "growing %s from %i to %i\n", HASHTABLE_STRING(HASHTABLE_NAME), h->hashtable_capacity, big->hashtable_capacity));
    for (i= 0;  i < h->hashtable_capacity;  ++i)
	if (h->hashtable_entries[i])
	    HASHTABLE(insert_entry)(big, h->hashtable_entries[i]);
    *h= *big;
    free(big);
    return h->hashtable_capacity;
}

HASHTABLE_ENTRY_T *HASHTABLE(find)(HASHTABLE_T *t, HASHTABLE_KEY_T key, unsigned l)
{
    unsigned h= HASHTABLE_HASH_FUNCTION(key, l) % t->hashtable_size;
    HASHTABLE_ENTRY_T *entry= t->hashtable_entries[h];
    HASHTABLE_DEBUG((stderr, "FIND\t%s\t%08x %p\n", key, h, entry));
    if (entry) 	/* search the chain at index h */
	for (;  entry;  entry= entry->hashtable_entry_next)
	    if ((l == entry->hashtable_entry_length) && HASHTABLE_KEY_COMPARE(key, entry->hashtable_entry_key))
		return entry;
    return 0;
}

/* undefine everything so that other files can include this one more than once with differnt types */

#undef HASHTABLE_KEEP_DEFINITIONS
#undef HASHTABLE_DEBUG
#undef HASHTABLE_INFO
#undef HASHTABLE_KEY_COPY
#undef HASHTABLE_KEY_COMPARE
#undef HASHTABLE_FRACTION
#undef HASHTABLE_FACTOR
#undef HASHTABLE_HASH_FUNCTION
#undef HASHTABLE_STRING2
#undef HASHTABLE_STRING
#undef HASHTABLE_CONCAT2
#undef HASHTABLE_CONCAT
#undef HASHTABLE
#undef HASHTABLE_NAME
#undef HASHTABLE_NAME_DEFINED
#undef HASHTABLE_ENTRY_NAME
#undef HASHTABLE_ENTRY_NAME_DEFINED
#undef HASHTABLE_KEY_T
#undef HASHTABLE_KEY_T_DEFINED
#undef HASHTABLE_T
#undef HASHTABLE_T_DEFINED
#undef HASHTABLE_ENTRY_T
#undef HASHTABLE_ENTRY_T_DEFINED
